@@ -4,7 +4,8 @@ mod _functiontrace_rs {
     use functiontrace_server::function_trace;
     use pyo3::prelude::*; // TODO: What's in here?
     use serde::Serialize;
-    use std::ffi::{c_void, OsString};
+    use std::borrow::Cow;
+    use std::ffi::{c_void, CStr, OsString};
     use std::io::Write;
     use std::os::unix::net::UnixStream;
     use std::sync::atomic::{AtomicBool, AtomicU32, Ordering};
@@ -46,11 +47,7 @@ mod _functiontrace_rs {
         /// A small [`Write`] implementation that allows us to write to an
         /// `mpack_writer_t`'s interior buffer while keeping its state consistent.
         fn write(&mut self, src: &[u8]) -> std::io::Result<usize> {
-            if !STARTED
-                .get()
-                .expect("Set in `set_config`")
-                .load(Ordering::Relaxed)
-            {
+            if !STARTED.load(Ordering::Relaxed) {
                 // We haven't started yet, so we shouldn't write anything out.  This avoids logging
                 // various internal setup things before we're ready.
                 return Ok(src.len());
@@ -140,18 +137,11 @@ mod _functiontrace_rs {
 
     /// True iff we've started tracing and haven't been marked as terminated.  When this is
     /// set, we're allowed to send messages to the profile generation server.
-    static STARTED: OnceLock<&'static AtomicBool> = OnceLock::new();
+    static STARTED: AtomicBool = AtomicBool::new(false);
 
     /// The `Fprofile_FunctionTrace` function defined in the C extension.  This is needed to finish
     /// moving setup to the Rust code.
-    static TRACEFUNC: OnceLock<
-        unsafe extern "C" fn(
-            *mut pyo3::ffi::PyObject,
-            *mut pyo3::ffi::PyFrameObject,
-            i32,
-            *mut pyo3::ffi::PyObject,
-        ) -> i32,
-    > = OnceLock::new();
+    static TRACEFUNC: OnceLock<pyo3::ffi::Py_tracefunc> = OnceLock::new();
 
     ////////////////////////////////////////////////////////////////////////////////////////////////
     // Rust <-> C Communication
@@ -176,16 +166,9 @@ mod _functiontrace_rs {
         Box::into_raw(api) as usize
     }
 
+    /// Called by the C module to pass the `c_functiontrace` handler to us.
     #[unsafe(no_mangle)]
-    extern "C" fn set_config(
-        started: *const AtomicBool,
-        functiontrace: unsafe extern "C" fn(
-            *mut pyo3::ffi::PyObject,
-            *mut pyo3::ffi::PyFrameObject,
-            i32,
-            *mut pyo3::ffi::PyObject,
-        ) -> i32,
-    ) -> libc::pthread_key_t {
+    extern "C" fn set_config(c_functiontrace: pyo3::ffi::Py_tracefunc) -> libc::pthread_key_t {
         // Allocate a Tss_Key for storing thread state
         let mut tss_key = 0;
         if unsafe {
@@ -196,16 +179,8 @@ mod _functiontrace_rs {
             panic!("Failed to create Tss_Key");
         }
 
-        STARTED
-            .set(unsafe {
-                started
-                    .as_ref()
-                    .expect("started is non-null and has a static lifetime")
-            })
-            .expect("set_config can only be called once");
-
         TRACEFUNC
-            .set(functiontrace)
+            .set(c_functiontrace)
             .expect("set_config can only be called once");
 
         TSS_KEY.store(tss_key, Ordering::Relaxed);
@@ -339,7 +314,6 @@ mod _functiontrace_rs {
         module: &Bound<'_, PyModule>,
         output_directory: OsString,
     ) -> Result<()> {
-        //fn trace_initialization(args: *mut pyo3::ffi::PyObject) -> Result<()> {
         // The name of the env var we use to notify subprocesses that there's a trace in progress.
         const BREADCRUMB: &str = "FUNCTIONTRACE_LIVE";
 
@@ -544,11 +518,8 @@ mod _functiontrace_rs {
         // Register our tracing functions with Python - both the normal one and the multithreaded
         // setup trampoline.
         unsafe {
-            // SAFETY: tracefunc is a valid function pointer to `Fprofile_FunctionTrace`.
-            pyo3::ffi::PyEval_SetProfile(
-                Some(*TRACEFUNC.get().expect("tracefunc is set in set_config")),
-                std::ptr::null_mut(),
-            )
+            // SAFETY: both parameters are of the proper type.
+            pyo3::ffi::PyEval_SetProfile(Some(functiontrace), std::ptr::null_mut())
         };
 
         Python::with_gil(|py| -> Result<()> {
@@ -568,10 +539,7 @@ mod _functiontrace_rs {
         crate::hooks::install().wrap_err("Failed to install Python hooks")?;
 
         // We're now fully setup and allowed to send messages.
-        STARTED
-            .get()
-            .expect("Set in `set_config`")
-            .store(true, Ordering::Relaxed);
+        STARTED.store(true, Ordering::Relaxed);
 
         Ok(())
     }
@@ -585,10 +553,7 @@ mod _functiontrace_rs {
         _kwargs: Option<&Bound<'_, pyo3::types::PyDict>>,
     ) -> Result<()> {
         // Stop profiling and don't allow further logging, then flush any remaining messages.
-        STARTED
-            .get()
-            .expect("module_configuration was set by `set_config`")
-            .store(false, Ordering::Relaxed);
+        STARTED.store(false, Ordering::Relaxed);
 
         unsafe {
             // PyEval_SetProfile(NULL, NULL) is the proper way to deregister a profiler.
@@ -723,11 +688,7 @@ mod _functiontrace_rs {
     fn thread_reset() -> Result<()> {
         let tss_key = TSS_KEY.load(Ordering::Relaxed);
 
-        if !STARTED
-            .get()
-            .expect("STARTED was set by `set_config`")
-            .load(Ordering::Relaxed)
-        {
+        if !STARTED.load(Ordering::Relaxed) {
             // We haven't actually started yet, but are for some reason being asked
             // to fork.  This appears to be OS dependent.
             return Ok(());
@@ -782,11 +743,8 @@ mod _functiontrace_rs {
 
         // Replace our setprofile() handler with the real one.
         unsafe {
-            // SAFETY: tracefunc is a valid function pointer to `Fprofile_FunctionTrace`.
-            pyo3::ffi::PyEval_SetProfile(
-                Some(*TRACEFUNC.get().expect("tracefunc is set in set_config")),
-                std::ptr::null_mut(),
-            )
+            // SAFETY: both parameters are of the proper type.
+            pyo3::ffi::PyEval_SetProfile(Some(functiontrace), std::ptr::null_mut())
         };
 
         // We previously called into `Fprofile_FunctionTrace` here to manually record this call,
@@ -854,6 +812,167 @@ mod _functiontrace_rs {
 
         Ok(())
     }
+
+    /// The core logging function for FunctionTrace.  Configured by `PyEval_SetProfile`, and used
+    /// to record all function calls/returns.
+    extern "C" fn functiontrace(
+        obj: *mut pyo3::ffi::PyObject,
+        frame: *mut pyo3::ffi::PyFrameObject,
+        what: i32,
+        arg: *mut pyo3::ffi::PyObject,
+    ) -> i32 {
+        let Some(mut writer) = thread_state() else {
+            // No reason to look at the event since we can't log it yet.
+            return 0;
+        };
+
+        if !STARTED.load(Ordering::Relaxed) {
+            // We aren't running (either starting up or shutting down), so shouldn't write
+            // anything.
+            return 0;
+        }
+
+        let time = trace_time();
+        let c_functiontrace = *TRACEFUNC.get().expect("tracefunc is set in set_config");
+
+        // Return the function name when we're in a C_* context.
+        let c_func_name = || {
+            let func = arg as *mut pyo3::ffi::PyCFunctionObject;
+
+            unsafe {
+                // SAFETY: In the `C_*` cases, `arg` is a non-NULL `*PyCFunctionObject`, which
+                // always contains a valid method.
+                let method = (*func).m_ml;
+                (*method).ml_name
+            }
+        };
+        // Given a C string sourced from Python, convert it to a Cow that can be serialized.
+        let cstr_to_cow = |ptr: *const i8| {
+            if !ptr.is_null() {
+                unsafe {
+                    // SAFETY: `ptr` is non-NULL and contains a valid C string.
+                    // TODO: Can this change underneath us?
+                    CStr::from_ptr(ptr)
+                }
+                .to_str()
+                .unwrap_or("<invalid string>")
+                .into()
+            } else {
+                Cow::Borrowed("<null>")
+            }
+        };
+
+        match what {
+            pyo3::ffi::PyTrace_CALL | pyo3::ffi::PyTrace_RETURN => {
+                // For now we delegate these to the C implementation, since we can't access the
+                // `PyCodeObject` fields that we require via Rust.
+                unsafe {
+                    // SAFETY: tracefunc is a valid function pointer to `Fprofile_FunctionTrace`.
+                    return c_functiontrace(obj, frame, what, arg);
+                }
+            }
+            pyo3::ffi::PyTrace_C_CALL => {
+                let func = unsafe {
+                    // SAFETY: In the `C_CALL` case, `arg` is guaranteed to be a non-NULL pointer
+                    // to a `PyCFunctionObject`.
+                    &*(arg as *mut pyo3::ffi::PyCFunctionObject)
+                };
+
+                let func_name = cstr_to_cow(c_func_name());
+
+                // We either belong to a module or are a method on a class.
+                if !func.m_module.is_null() {
+                    // NOTE: _Typically_ `m_module` is a PyModule, but there are no requirements on
+                    // what type it needs to be!  In practice, we only see PyModules and PyStrings.
+                    Python::with_gil(|py| {
+                        let m_module = unsafe {
+                            // SAFETY: m_module is a non-NULL pointer to a valid Python object.
+                            Bound::from_borrowed_ptr(py, func.m_module)
+                        };
+
+                        if let Ok(module) = m_module.downcast::<pyo3::types::PyModule>() {
+                            // This is a PyModule, so get its name if possible.
+                            let name = module.name();
+                            let module_name = if let Ok(name) = &name {
+                                name.to_cow().unwrap_or(Cow::Borrowed("<invalid string>"))
+                            } else {
+                                Cow::Borrowed("<unnamed module>")
+                            };
+
+                            function_trace::FunctionTrace::NativeCall {
+                                module_name,
+                                func_name,
+                                time,
+                            }
+                            .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
+                        } else {
+                            let module_name =
+                                if let Ok(module) = m_module.downcast::<pyo3::types::PyString>() {
+                                    module.to_cow().unwrap_or(Cow::Borrowed("<invalid string>"))
+                                } else {
+                                    // TODO: tracing log this
+                                    Cow::Borrowed("<unknown module>")
+                                };
+                            function_trace::FunctionTrace::NativeCall {
+                                module_name,
+                                func_name,
+                                time,
+                            }
+                            .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
+                        }
+                    })
+                } else {
+                    // We don't belong to a module, so must be a class method.
+                    let module_name = if !func.m_self.is_null() {
+                        cstr_to_cow(unsafe {
+                            // SAFETY: Since `m_self` is non-NULL, it must point to a valid
+                            // PyObject, which is required to have a live and fully initialized
+                            // `ob_type` member.
+                            (*(*func.m_self).ob_type).tp_name
+                        })
+                    } else {
+                        Cow::Borrowed("<unknown module>")
+                    };
+
+                    function_trace::FunctionTrace::NativeCall {
+                        module_name,
+                        func_name,
+                        time,
+                    }
+                    .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
+                }
+                .expect("Failed to emit NativeCall message");
+            }
+            pyo3::ffi::PyTrace_C_RETURN | pyo3::ffi::PyTrace_C_EXCEPTION => {
+                if what == pyo3::ffi::PyTrace_C_EXCEPTION {
+                    // C exceptions aren't exposed in a particularly interesting fashion, since they're
+                    // done using the `PyErr_*` functions plus a NULL return value.
+                    // However, if a `C_EXCEPTION` event is sent, we won't receive the corresponding
+                    // `C_RETURN` event, so in practice we treat both of these events the same.
+                    //
+                    // TODO: Should we try exposing information about them?  We'd probably need to
+                    // change the `Exception` format in `functiontrace-server`.
+                }
+
+                function_trace::FunctionTrace::NativeReturn {
+                    func_name: cstr_to_cow(c_func_name()),
+                    time,
+                }
+                .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
+                .expect("Failed to emit NativeReturn message");
+            }
+            pyo3::ffi::PyTrace_EXCEPTION => {
+                // TODO: Actually handle exceptions and do something interesting.
+                // We unfortunately can't get these from `PyEval_SetProfile`, so we'd need to use a
+                // different method to capture them.
+            }
+            _ => {
+                // TODO: this shouldn't be possible - tracing log it so we can handle later.
+            }
+        };
+
+        0
+    }
 }
 
 /// Allocation hooks and helpers for tracing Python memory allocator operations.
@@ -865,14 +984,12 @@ mod allocation_wrappers {
     /// Log the given allocation details.
     fn allocation_log<F: FnOnce() -> function_trace::AllocationDetails>(msg: F) {
         if let Some(mut writer) = crate::_functiontrace_rs::thread_state() {
-            let allocation = function_trace::FunctionTrace::Allocation {
+            function_trace::FunctionTrace::Allocation {
                 time: crate::_functiontrace_rs::trace_time(),
                 details: msg(),
-            };
-
-            allocation
-                .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
-                .expect("Failed to emit Allocation message");
+            }
+            .serialize(&mut rmp_serde::encode::Serializer::new(&mut writer))
+            .expect("Failed to emit Allocation message");
         }
     }
 
